;; auto-generated

(defconst mu4e-mu-version "1.12.7"
  "Required mu binary version; mu4e's version must agree with this.")

(defconst mu4e-doc-dir "/usr/local/share/doc/mu"
  "Mu4e's data-dir.")

(provide 'mu4e-config)
