;; -*- no-byte-compile: t; -*-
(define-package "mu4e" "1.12.7"
  "part of mu4e, the mu mail user agent"
  '((emacs "26.3"))
  :authors '(("Dirk-Jan C. Binnema" . "djcb@djcbsoftware.nl"))
  :maintainer '("Dirk-Jan C. Binnema" . "djcb@djcbsoftware.nl")
  :keywords '("email"))
